#pragma once
#ifndef MESSMER_CPPUTILS_SYSTEM_GETTOTALMEMORY_H
#define MESSMER_CPPUTILS_SYSTEM_GETTOTALMEMORY_H

#include <cstdint>

namespace cpputils {
    namespace system {
        uint64_t get_total_memory();
    }
}

#endif
