// nr.h - originally written and placed in the public domain by Wei Dai

/// \file nr.h
/// \brief Classes for Nyberg-Rueppel signature scheme

#ifndef CRYPTOPP_NR_H
#define CRYPTOPP_NR_H

#include "gfpcrypt.h"

#endif
